/**
 * BNSim is an open-source, parallel, stochastic, and multi-scale modeling
 * platform which integrates various simulation algorithms, together with
 * genetic circuits and chemotaxis pathway models in a 3D environment.
 *
 * BNSim is developed by CMU SLD Group, and released under GPLv2 license
 * Please check http://www.ece.cmu.edu/~sld/bnsim for more information
 *
 * BNSim is developed using C++ and pthread under Linux and Mac OS
 *
 * If you use it for your work, please cite
 * Wei, Guopeng, Paul Bogdan, and Radu Marculescu. "Efficient Modeling and
 * Simulation of Bacteria-Based Nanonetworks with BNSim." Selected Areas in
 * Communications, IEEE Journal on 31.12 (2013): 868-878.
 *
 * @author      Guopeng (Daniel) Wei  danielwei@cmu.edu
 * @version     2.5
 * @since       1.0
 */



#include <cstdlib>
#include "../core/core.h"


using namespace std;
using namespace BNSim;

QuorumSensingSystem::QuorumSensingSystem(Bacteria * host) {

	A1 = 0;
	C1 = 0;
	S = 0;
	R1 = 0;
	R1A1 = 0;

	R1QSI1 = 0;

	// rhl
	A2 = 0;
	C2 = 0;
	R2 = 0;
	R2A2 = 0;

	resistantMutant = false;

	// define rate constants
	kA1 = 2e-3;
	kA2 = 2e-3;
	kR1 = 2e-3;
	kR2 = 2e-3;
	kS = 2e-3;
	CA1 = 1e-4;
	CA2 = 1e-4;
	CR1 =  1e-4;
	CR2 =  1e-4;
	CS =  1e-4;

	KL1 = 1e-6;
	KL2 = 1e-6;
	KL3 = 1e-6;
	KL4 = 1e-6;
	KS1 = 1e-6;
	KS = 1e-6;
	KR1 = 1e-6;

    this->activated = false;
    
    int i = 0;
    while(i++!=12) {
        k[i] = 0.1;
    }

    k[0] = k[7] = 0.01;
    k[3] = k[10] = k[6] = 0.01;

    this->host = host;
}

double QuorumSensingSystem::randomUniform() {
	return ((double) rand() / (RAND_MAX));
}

void QuorumSensingSystem::tick() {

	double deltaA1 = (CA1 + kA1*C1/(KL1+C1)- k[0]*A1 - k[1]*A1*R1 + k[2]*R1A1)*Sim::param.timeStep;
	double deltaR1 = (CR1 + kR1*C1/(KR1+C1) - k[3]*R1 - k[1]*A1*R1 + k[2]*R1A1)*Sim::param.timeStep;
	double deltaR1A1 = (k[1]*A1*R1 - k[2]*R1A1 -2*k[4]*R1A1*R1A1 + 2*k[5]*C1)*Sim::param.timeStep;
	double deltaC1 = (k[4]*R1A1*R1A1 - k[5]*C1)*Sim::param.timeStep;

	A1 += (deltaA1 + 0.1*deltaA1*(randomUniform()-0.5));
	C1 += (deltaC1 + 0.1*deltaC1*(randomUniform()-0.5));
	R1 += (deltaR1 + 0.1*deltaR1*(randomUniform()-0.5));
	R1A1 += (deltaR1A1 + 0.1*deltaR1A1*(randomUniform()-0.5));
    
	double deltaA2 = (CA2 + kA2*C2/((KL2+C2)) - k[7]*A2 - k[8]*A2*R2 + k[9]*R2A2)*Sim::param.timeStep;
	double deltaR2 = (CR2 + kR2*C2*C1/((KL3+C2)*(KL4+C1)) - k[10]*R2 - k[8]*A2*R2 + k[9]*R2A2)*Sim::param.timeStep;
	double deltaR2A2 = (k[8]*A2*R2 - k[9]*R2A2 -2*k[11]*R2A2*R2A2 + 2*k[12]*C2)*Sim::param.timeStep;
	double deltaC2 = (k[11]*R2A2*R2A2 - k[12]*C2)*Sim::param.timeStep;

	A2 += (deltaA2 + 0.1*deltaA2*(randomUniform()-0.5));
	C2 += (deltaC2 + 0.1*deltaC2*(randomUniform()-0.5));
	R2 += (deltaR2 + 0.1*deltaR2*(randomUniform()-0.5));
	R2A2 += (deltaR2A2 + 0.1*deltaR2A2*(randomUniform()-0.5));
    
	if(host!=NULL) {
        
        
        EnvCube* Cube = Sim::univ->getCube(host->getCubeLoc());
        
        // Now let's exchange molecules with extracellular space
        double D = 0.1;

        double ex = Cube->chemicals[0];
        double delta = (A1 - ex)*D*Sim::param.timeStep;
        A1 -= delta;
        Cube->chemicals[0] += delta;   //mM
	}
}


